"""Module containing functions related to different rotation representations."""
import numpy as np
from vecmaths.utils import prt
from vecmaths import vectors


def _vecpair2rotmat_stack(veca, vecb):
    """
    Generate stacks of rotation matrices which rotate from one set of column
    vector directions to another.

    Parameters
    ----------
    veca : ndarray
        Stacks of column three-vectors. Must have innermost shape of (3, 1).
    vecb : ndarray
        Stacks of column three-vectors. Must have innermost shape of (3, 1).

    Returns
    -------
    rot_mat : ndarray
        Stacks of rotation matrices with innermost shape of (3, 3).

    """

    if veca.shape != vecb.shape:
        raise ValueError('Shapes of `veca` and `vecb` do not '
                         'match: {} and {}'.format(veca.shape, vecb.shape))

    if veca.shape[-2:] != (3, 1):
        raise ValueError('Innermost shapes of `veca` and `vecb` '
                         'must be (3, 1), but is {}.'.format(veca.shape[-2:]))

    # Normalise
    veca = veca / np.linalg.norm(veca, axis=-2)[..., None]
    vecb = vecb / np.linalg.norm(vecb, axis=-2)[..., None]

    # Cross product matrix
    vecx = np.cross(veca, vecb, axis=-2)
    xprod = vectors.cross_prod_mat(vecx)

    # Note axis `j` has size 1:
    dot = np.einsum('...ij,...ij->...j', veca, vecb)[..., None]

    # Find which vector pairs are anti-parallel:
    neg = np.isclose(dot, -1)
    neg_idx_full = np.where(neg)
    not_neg = np.logical_not(neg)
    not_neg_idx_full = np.where(not_neg)

    # Vector indices rather than component indices:
    neg_idx = neg_idx_full[:-2]
    not_neg_idx = not_neg_idx_full[:-2]

    # Find rotation matrices for anti-parallel vector pairs:
    veca_neg = veca[neg_idx]
    veca_perp = vectors.perpendicular(veca_neg, axis=-2)
    rot_mat_neg = axang2rotmat(veca_perp, 180, degrees=True, axis=-2)

    rot_mat = np.zeros_like(xprod)
    rot_mat[neg_idx] = rot_mat_neg

    if not_neg_idx_full[0].size:

        # Find rotation matrices for remaining vector pairs:
        xprod_nn = xprod[not_neg_idx]
        dot_nn = dot[not_neg_idx]

        rot_mat[not_neg_idx] = np.eye(3) + xprod_nn
        rot_mat[not_neg_idx] += (xprod_nn @ xprod_nn) / (1 + dot_nn)

    return rot_mat


def vecpair2rotmat(veca, vecb, axis=0):
    """
    Generate stacks of rotation matrices which rotate from one set of vector
    direction to another.

    Parameters
    ----------
    veca : ndarray
        Array of vectors whose directions are considered the initial
        directions. Must have the same shape as `vecb`.
    vecb : ndarray
        Array of vectors whose directions are considered the final directions.
        Must have the same shape as `veca`.
    axis : int
        Axis defining the vectors.

    Returns
    -------
    rot_mat : ndarray
        Stacks of rotation matrices with innermost shape of (3, 3).

    Notes
    -----
    1.  Size-1 dimensions whose axis positions are greater than `axis` are
        squeezed out.

    """

    if veca.shape != vecb.shape:
        raise ValueError('Shapes of `veca` and `vecb` do not '
                         'match: {} and {}'.format(veca.shape, vecb.shape))

    if veca.shape[axis] != 3:
        raise ValueError('Size of dimension `axis` ({}) along `veca` and '
                         '`vecb` must be 3, but is ''{}.'.format(
                             axis, veca.shape[axis]))

    # Squeeze out size-1 dimensions after `axis`
    inner_ax_shape = np.array(list(veca.shape[axis:]))
    squeeze_axes = tuple(np.where(inner_ax_shape == 1)[0] + axis)
    veca = veca.squeeze(axis=squeeze_axes).swapaxes(-1, axis)
    vecb = vecb.squeeze(axis=squeeze_axes).swapaxes(-1, axis)

    # Ensure the innermost shape is (3, 1)
    veca = veca[..., None]
    vecb = vecb[..., None]

    return _vecpair2rotmat_stack(veca, vecb)


def axang2rotmat(rot_ax, ang, axis=0, ndim_outer=0, degrees=False):
    """
    Find the rotation matrices from axis-angle representations.

    Generated rotation matrices act on column vectors by pre-multiplication.
    Where multiple rotation matrices are generated, the returned array has
    innermost dimensions of (3, 3) corresponding to each rotation matrix.

    Parameters
    ----------
    rot_ax : ndarray
        One or more rotation axes. See Notes for details on allowed shapes.
    ang : float or list of float or ndarray
        One or more rotation angles. See Notes for details on allowed shapes.
    axis : int
        The axis of `rot_ax` corresponding to each rotation axis. This axis
        should have length three. By default, set to 0.
    ndim_outer : int, optional
        Determines the number of outer dimensions. The first `ndim_outer`
        dimensions of both `ax` and `ang` must have the same size. By
        default, set to 0.
    degrees : bool, optional
        If True, `ang` is interpreted as degrees, else radians. False by
        default.

    Returns
    -------
    rot_mat : ndarray of inner shape (3, 3)

    Notes
    -----
    1.  Uses the matrix form of the Rodrigues' rotation formula.

    2.  The returned rotation matrix array has an inner shape determined
        according to the input inner shapes of `rot_ax` and `angle` and the
        value of `axis`:

        `rot_ax` shape  | `axis` | `ang` shape | `rot_mat` shape
        ----------------|--------|-------------|----------------
        (3,)            | 0      | ()          | (3, 3)
        (3, 1)          | 0      |             |
        (1, 3)          | 1      |             |
        ----------------|--------|-------------|----------------
        (3,)            | 0      | (M,)        | (M, 3, 3)
        (3, 1)          | 0      |             |
        (1, 3)          | 1      |             |
        ----------------|--------|-------------|----------------
        (N, 3)          | 1      | ()          | (N, 3, 3)
        (N, 3, 1)       | 1      |             |
        (3, N)          | 0      |             |

        Additional dimensions of size 1 within the inner shape are squeezed
        out.

        In addition, an outer shape `P` (i.e. shape (p1, p2, ...)) is allowed,
        in which case `ndim_outer` should be set to the number of dimensions in
        the outer shape (i.e. `len(P)`).

    """

    ang = np.array(ang)
    inner_ang_shp = ang.shape[ndim_outer:]

    if degrees:
        ang = np.deg2rad(ang)

    # Convert to a positive axis index:
    if axis < 0:
        axis = rot_ax.ndim + axis

    if axis >= rot_ax.ndim:
        raise ValueError('`axis` must be less than the number of '
                         'dimensions in `rot_ax`.')

    if axis < ndim_outer:
        raise ValueError('`axis` cannot be less than `ndim_outer.')

    if rot_ax.shape[axis] != 3:
        raise ValueError('Specified axis of `rot_ax` ({}) must have '
                         'size 3.'.format(axis))

    outer_rot_ax_shp = rot_ax.shape[:ndim_outer]
    outer_ang_shp = ang.shape[:ndim_outer]

    if outer_rot_ax_shp != outer_ang_shp:
        msg = (
            'The first `ndim_outer` ({}) dimensions of `rot_ax` and `ang` '
            'must match, but shapes are: {} and {}.'.format(
                ndim_outer, outer_rot_ax_shp, outer_ang_shp)
        )
        raise ValueError(msg)

    # Swap axes to get the rotation axis along the final dimension
    rot_ax = np.swapaxes(rot_ax, axis, -1)

    # Squeeze out any size-one dimensions from the inner shape:
    inner_ax_dim = rot_ax.ndim - ndim_outer
    inner_ax_shape = np.array(list(rot_ax.shape[-inner_ax_dim:]))
    squeeze_axes = tuple(np.where(inner_ax_shape == 1)[0] + ndim_outer)
    rot_ax = rot_ax.squeeze(axis=squeeze_axes)

    # Ensure the innermost `rot_ax` shape is (3, 1)
    rot_ax = rot_ax[..., None]

    inner_ang_dim = ang.ndim - ndim_outer
    inner_ax_dim = rot_ax.ndim - ndim_outer
    outer_ax_shp = rot_ax.shape[:-2]

    if inner_ang_dim not in [0, 1]:
        raise ValueError('`ang` must have inner dimension of zero or one.')

    if inner_ang_dim + inner_ax_dim > 3:
        raise ValueError('Combined number of inner dimensions of `ax` and '
                         '`ang` should be two or three.')

    # Check final two dimensions of `ax` are column three-vectors:
    if rot_ax.shape[-2:] != (3, 1):
        raise ValueError('Inner shape of `ax` must be (N, 3, 1) or (3, 1) or '
                         '(3,).')

    # Normalise axis to unit vector:
    rot_ax_norm = np.linalg.norm(rot_ax, axis=-2)[..., None]
    rot_ax = rot_ax / rot_ax_norm

    ang = ang[..., None, None]

    if inner_ang_dim == 1:
        # Broadcast axis to account for multiple angles
        ax_exp = np.expand_dims(rot_ax, ndim_outer)
        ax_newshp = outer_ax_shp + inner_ang_shp + (3, 1)
        ang_newshp = outer_ax_shp + inner_ang_shp + (1, 1,)
        rot_ax = np.broadcast_to(ax_exp, ax_newshp)
        ang = np.broadcast_to(ang, ang_newshp)

    xprod = vectors.cross_prod_mat(rot_ax)
    rot_mat = np.eye(3) + np.sin(ang) * xprod
    rot_mat += (1 - np.cos(ang)) * (xprod @ xprod)

    return rot_mat


def rotmat2axang(rot_mat, degrees=False, tol=1e-10):
    """
    Decompose rotation matrices into axis-angle representations.

    Parameters
    ----------
    rot_mat : ndarray
        Stacks of rotation matrices with innermost shape (3, 3).
    degrees : bool, optional
        If True, angles are returned in degrees, otherwise in radians. Default
        is False.

    Returns
    -------
    rot_ax : ndarray
        Stacks of column vectors representing rotation axes for each rotation
        matrix. If input `rot_mat` has shape (N, M, ..., 3, 3), `rot_ax` will
        have shape (N, M, ..., 3, 1).
    angle : ndarray
        Rotation angles. If input `rot_mat` has shape (N, M, ..., 3, 3),
        `angle` will have shape (N, M, ...).

    Notes
    -----
    Following similar function in Matlab from here:
    https://github.com/marcdegraef/3Drotations/blob/master/src/MatLab/om2ax.m

    TODO:
    - Understand `P` factor in  http://doi.org/10.1088/0965-0393/23/8/083501
      and apply here if necessary.

    """

    # Check dimensions
    if rot_mat.shape[-2:] != (3, 3):
        raise ValueError('`rot_mat` must have inner dimensions (3, 3).')

    single_mat = False
    if rot_mat.shape == (3, 3):
        single_mat = True
        rot_mat = rot_mat[None]

    trc = np.trace(rot_mat, axis1=-2, axis2=-1)
    angle = np.arccos(np.clip(0.5 * (trc - 1), -1, 1))

    if degrees:
        angle = np.rad2deg(angle)

    ax_shp = rot_mat.shape[:-2] + (3,)
    rot_ax = np.ones(ax_shp) * np.nan

    ang_zero = np.isclose(angle, 0.0)
    ang_zero_idx = np.where(ang_zero)

    if ang_zero_idx[0].size:
        rot_ax[ang_zero_idx] = [0, 0, 1]

    # Find eigenvalues, eigenvectors for `rot_mat`
    eigval, eigvec = np.linalg.eig(rot_mat)

    # Get index of eigenvalue which is 1 + 0i
    eig_cond = np.logical_and(
        abs(np.real(eigval) - 1) < tol,
        abs(np.imag(eigval) < tol)
    )
    eig_cond_sum = np.sum(eig_cond, axis=-1)

    # Find indices where rotation axes cannot be extracted:
    no_ax = np.logical_and(
        eig_cond_sum != 1,
        np.logical_not(ang_zero)
    )
    no_ax_idx = np.where(no_ax)

    if no_ax_idx[0].size:
        no_ax_all_idx = (np.vstack(no_ax_idx).T).tolist()
        msg = ('Not exactly one eigenvector with eigenvalue of 1 found for '
               'rotation matrices at indices: {}'.format(no_ax_all_idx))
        raise ValueError(msg)

    eig_cond[ang_zero] = False
    eig_cond_idx = np.where(eig_cond)

    # Swap final axes of eigenvectors to allow indexing column vectors
    axes_comp = eigvec.swapaxes(-2, -1)[eig_cond_idx]

    # Set the axes to eigenvector with eigenvalue == 1
    rot_ax[eig_cond_idx[:-1]] = np.real(axes_comp)

    # Sort out signs of axes components:
    ax_sign_calc = np.ones_like(rot_ax)
    ax_sign_calc[..., 0] = rot_mat[..., 2, 1] - rot_mat[..., 1, 2]
    ax_sign_calc[..., 1] = rot_mat[..., 0, 2] - rot_mat[..., 2, 0]
    ax_sign_calc[..., 2] = rot_mat[..., 1, 0] - rot_mat[..., 0, 1]

    ax_sign_nz = np.logical_not(np.isclose(ax_sign_calc, 0))
    ax_sign = np.sign(ax_sign_calc[ax_sign_nz])
    rot_ax[ax_sign_nz] = np.abs(rot_ax[ax_sign_nz]) * ax_sign

    # Return axes as stacks of column vectors
    rot_ax = rot_ax[..., None]

    if single_mat:
        rot_ax = rot_ax[0]
        angle = angle[0]

    return (rot_ax, angle)


def euler2rotmat():
    pass


def rotmat2euler():
    pass


def align_xy(box):
    """Align a parallelepiped such that the first edge vector is along x and
    the second is in the xy-plane.

    Parameters
    ----------
    box : ndarray of shape (3, 3)
        Array of column vectors representing edge vectors of a parallelepiped.

    Returns
    -------
    aligned_box : ndarray of shape (3, 3)
        Array of column vectors representing edge vectors of rotated
        parallelepiped.

    """

    veca = box[:, 0]
    vecb = box[:, 1]
    vecc = box[:, 2]

    amag, bmag, cmag = [np.linalg.norm(i) for i in [veca, vecb, vecc]]

    cos_gamma = vectors.vecpair_cos(veca, vecb)
    sin_gamma = vectors.vecpair_sin(veca, vecb)
    cos_beta = vectors.vecpair_cos(vecc, veca)
    bc_dot = np.einsum('i,i->', vecb, vecc)

    b_x = bmag * cos_gamma
    b_y = bmag * sin_gamma
    c_x = cmag * cos_beta
    c_y = (bc_dot - (b_x * c_x)) / b_y

    veca_new = np.array([amag, 0, 0])
    vecb_new = np.array([b_x, b_y, 0])
    vecc_new = np.array([c_x, c_y, np.sqrt(cmag**2 - c_x**2 - c_y**2)])

    aligned_box = np.vstack([veca_new, vecb_new, vecc_new]).T

    return aligned_box


def get_random_rotation_matrix(shape=None, method='orthonormal_basis', max_angle_deg=None):
    """Generate a stack of random rotation matrices."""

    allowed_methods = ['orthonormal_basis', 'axis_angle']
    if method not in allowed_methods:
        raise ValueError('`method` must be one of: {}'.format(allowed_methods))

    if max_angle_deg is not None and method != 'axis_angle':
        msg = '`max_angle_deg` must only be specified for `method=orthonormal_basis`.'
        raise ValueError(msg)

    if not shape:
        shape = (1,)

    single = False
    if shape == (1,):
        single = True

    if method == 'orthonormal_basis':

        # Task is to form sets of orthonormal bases in 3D

        # Choose the first vectors randomly:
        vecs_ab_shp = shape + (3, 1)
        vecs_a = np.random.random(vecs_ab_shp)

        # Get vectors perpendicular to `vecs_a`:
        vecs_b = vectors.perpendicular(vecs_a, axis=-2)

        # Get vectors perpendicular to `vecs_a` and `vecs_b`:
        vecs_c = np.cross(vecs_a, vecs_b, axisa=-2, axisb=-2, axisc=-2)

        # Concatenate and normalise:
        rot = np.concatenate([vecs_a, vecs_b, vecs_c], axis=-1)
        rot = vectors.normalise(rot, axis=-2)

    elif method == 'axis_angle':

        if max_angle_deg is None:
            max_angle_deg = np.rad2deg(2 * np.pi)

        # Form vectors by choosing two random spherical polar coordinates:

        u_1 = np.random.random(shape)
        u_2 = np.random.random(shape)
        theta = np.arccos((2 * u_1) - 1)
        phi = 2 * np.pi * u_2
        vec = np.concatenate([
            (np.sin(theta) * np.cos(phi))[..., None],
            (np.sin(theta) * np.sin(phi))[..., None],
            (np.cos(theta))[..., None],
        ], axis=-1)

        # Rotation angles (magnitude of rotation about above vectors):
        angle = np.random.random(shape) * np.deg2rad(max_angle_deg)

        # Combine to form stacks of rotation matrices:
        rot = axang2rotmat(vec, angle, axis=-1, degrees=False, ndim_outer=len(shape))

    if single:
        rot = rot[0]

    return rot



"""Module containing functions related to general vector operations.

TODO: use `normalise` throughout this module.

"""
from itertools import permutations, combinations
import numpy as np


def normalise(vecs, axis=-1):
    """Normalise an n-dimensional array of vectors.

    Only non-zero vectors are normalised.

    Parameters
    ----------
    vecs : ndarray of innermost shape (3,)
        The innermost axis defines the three vector.
    axis : int, optional
        The axis which defines the vectors. Set to the last 
        axis (-1) by default.

    Returns
    -------
    vecs_normd : ndarray with same shape as input array

    """
    if axis < 0:
        axis = len(vecs.shape) + axis

    norm = np.linalg.norm(vecs, axis=axis)

    vecs_shp_pre_ax = vecs.shape[:axis]
    vecs_shp_post_ax = vecs.shape[axis + 1:]
    norm_shpe = vecs_shp_pre_ax + (1,) + vecs_shp_post_ax

    norm_rs = norm.reshape(norm_shpe)
    norm_rs_rep = np.repeat(norm_rs, vecs.shape[axis], axis=axis)
    norm_non_zero = np.logical_not(np.isclose(norm_rs_rep, 0))

    vecs_normd = np.copy(vecs)
    vecs_normd[norm_non_zero] = (
        vecs_normd[norm_non_zero] / norm_rs_rep[norm_non_zero]
    )

    return vecs_normd


def validate_vecpairs(veca, vecb, axis):
    """Validate the input form of vector pair functions."""

    if veca.shape != vecb.shape:
        raise ValueError('`veca` and `vecb` must have the same shape, but have'
                         ' shapes: {} and {}.'.format(veca.shape, vecb.shape))

    if axis >= veca.ndim:
        raise ValueError('`veca` and `vecb` have {} dimensions, but axis '
                         'is: {}.'.format(veca.ndim, axis))

    if veca.shape[axis] != 3:
        raise ValueError('The `axis` ({}) dimension of `veca` and `vecb` must '
                         'define the axis along which the three-vectors are '
                         'defined and must have length 3, but have '
                         'length {}.'.format(axis, veca.shape[axis]))


def cross_prod_mat(axis):
    """
    Find the skew-symmetric matrix which allows representing the
    cross product with matrix multiplication.

    Parameters
    ----------
    ax : ndarray

    """

    if axis.ndim == 1:
        axis = axis[:, None]

    if axis.shape[-2:] != (3, 1):
        raise ValueError('Inner shape of `ax` must be (N, 3, 1)'
                         ' or (3, 1) or (3,).')

    ret_shape = list(axis.shape)
    ret_shape[-1] = 3
    ret_shape = tuple(ret_shape)

    ret = np.zeros(ret_shape)

    ret[..., 0, 1] = -axis[..., 2, 0]
    ret[..., 0, 2] = axis[..., 1, 0]
    ret[..., 1, 0] = axis[..., 2, 0]
    ret[..., 1, 2] = -axis[..., 0, 0]
    ret[..., 2, 0] = -axis[..., 1, 0]
    ret[..., 2, 1] = axis[..., 0, 0]

    return ret


def perpendicular(vec, axis=-1):
    """
    Get 3-vectors perpendicular to a given set of 3-vectors.

    Parameters
    ----------
    vec : ndarray
        Array of vectors.
    axis : int, optional
        The axis along which the 3-vectors are defined.

    Returns
    -------
    perp_vec : ndarray
        Array of the same shape as input array `vec`, where each output vector
        is perpendicular to its corresponding input vector.

    """

    if vec.shape[axis] != 3:
        raise ValueError('Size of dimension `axis` ({}) along `vec` must be 3,'
                         ' but is {}.'.format(axis, vec.shape[axis]))

    # Reshape:
    vec = np.swapaxes(vec, -1, axis)

    # Return array will have the same shape as input
    perp_vec = np.ones_like(vec) * np.nan

    # Find where first component magnitudes are larger than last:
    a_gt_c = np.abs(vec[..., 0]) > np.abs(vec[..., 2])
    a_notgt_c = np.logical_not(a_gt_c)

    # Make bool index arrays
    a_gt_c_0 = np.zeros_like(perp_vec, dtype=bool)
    a_gt_c_0[..., 0] = a_gt_c
    a_gt_c_1 = np.roll(a_gt_c_0, shift=1, axis=-1)
    a_gt_c_2 = np.roll(a_gt_c_1, shift=1, axis=-1)

    a_notgt_c_0 = np.zeros_like(perp_vec, dtype=bool)
    a_notgt_c_0[..., 0] = a_notgt_c
    a_notgt_c_1 = np.roll(a_notgt_c_0, shift=1, axis=-1)
    a_notgt_c_2 = np.roll(a_notgt_c_1, shift=1, axis=-1)

    # Set each component of the output vectors:
    perp_vec[a_gt_c_0] = vec[a_gt_c][..., 2]
    perp_vec[a_gt_c_1] = 0
    perp_vec[a_gt_c_2] = -vec[a_gt_c][..., 0]

    perp_vec[a_notgt_c_0] = 0
    perp_vec[a_notgt_c_1] = vec[a_notgt_c][..., 2]
    perp_vec[a_notgt_c_2] = -vec[a_notgt_c][..., 1]

    # Reshape to original shape:
    perp_vec = np.swapaxes(perp_vec, -1, axis)

    return perp_vec


def vecpair_angle(veca, vecb, ref=None, axis=-1, degrees=False):
    """Find the signed angles between a set of vectors and another.

    Parameters
    ----------
    veca : ndarray
        The first set of vectors.
    vecb : ndarray
        The second set of vectors.
    axis : int, optional
        The axis along which each 3-vector is defined. By default, the last
        axis.
    ref : ndarray of size 3 or None, optional
        Reference vector used to determine which +/- axis to use for the
        rotation. If ndarray, must have size 3. For each vector pair, the
        rotation axis closest to the reference vector will be chosen. If None,
        the reference vector is taken as the cross product of each vector pair,
        i.e. all returned angles will be positive. By default, None.
    degrees : bool, optional
        If True returned angles are in degrees, else in radians. By default,
        False.

    Returns
    -------
    theta : ndarray

    """

    validate_vecpairs(veca, vecb, axis)

    # Reshape
    veca = np.swapaxes(veca, -1, axis)
    vecb = np.swapaxes(vecb, -1, axis)

    vecx = np.cross(veca, vecb, axis=-1)
    vecx_normd = normalise(vecx)

    if ref is not None:
        ref = np.squeeze(ref)
        ref_normd = ref / np.linalg.norm(ref, axis=-1)
        ref_dot = np.einsum('i,...i->...', ref_normd, vecx_normd)
        vecx_normd[ref_dot < 0] *= -1

    dot_spec = '...i,...i->...'
    dot = np.einsum(dot_spec, veca, vecb)
    tri_prod = np.einsum(dot_spec, vecx, vecx_normd)
    theta = np.arctan2(tri_prod, dot)

    if degrees:
        theta = np.rad2deg(theta)

    return theta


def vecpair_cos(veca, vecb, axis=-1):
    """Find the cosines between a set of vectors and another.

    Parameters
    ----------
    veca : ndarray
        The first set of vectors.
    vecb : ndarray
        The second set of vectors.
    axis : int, optional
        The axis along which each 3-vector is defined. By default, the last
        axis.

    Returns
    -------
    cosines : ndarray
        Array with the same shape as input vectors except for the removed
        `axis` dimension.

    """
    validate_vecpairs(veca, vecb, axis)

    # Reshape
    veca = np.swapaxes(veca, -1, axis)
    vecb = np.swapaxes(vecb, -1, axis)

    veca_normd = veca / np.linalg.norm(veca, axis=-1)[..., None]
    vecb_normd = vecb / np.linalg.norm(vecb, axis=-1)[..., None]

    cosines = np.einsum('...i,...i->...', veca_normd, vecb_normd)[..., None]

    # Reshape to original axis order
    cosines = np.swapaxes(cosines, -1, axis).squeeze(axis=axis)

    return cosines


def vecpair_sin(veca, vecb, axis=-1):
    """Find the sines between a set of vectors and another.

    Parameters
    ----------
    veca : ndarray
        The first set of vectors.
    vecb : ndarray
        The second set of vectors.
    axis : int, optional
        The axis along which each 3-vector is defined. By default, the last
        axis.

    Returns
    -------
    sines : ndarray
        Array with the same shape as input vectors except for the removed
        `axis` dimension.

    """

    validate_vecpairs(veca, vecb, axis)

    # Reshape
    veca = np.swapaxes(veca, -1, axis)
    vecb = np.swapaxes(vecb, -1, axis)

    veca_normd = veca / np.linalg.norm(veca, axis=-1)[..., None]
    vecb_normd = vecb / np.linalg.norm(vecb, axis=-1)[..., None]

    xprod = np.cross(veca_normd, vecb_normd, axis=-1)

    # Reshape to original axis order
    xprod = np.swapaxes(xprod, -1, axis)
    sines = np.linalg.norm(xprod, axis=axis)

    return sines


def find_positive_int_vecs(search_size, dim=3):
    """
    Find arbitrary-dimension positive integer vectors which are
    non-collinear whose components are less than or equal to a
    given search size. Vectors with zero components are not included.

    Non-collinear here means no two vectors are related by a scaling factor.

    Parameters
    ----------
    search_size : int
        Positive integer which is the maximum vector component
    dim : int
        Dimension of vectors to search.

    Returns
    -------
    ndarray of shape (N, `dim`)

    """

    # Generate trial vectors as a grid of integer vectors
    search_ints = np.arange(1, search_size + 1)
    search_grid = np.meshgrid(*(search_ints,) * dim)
    trials = np.vstack(search_grid).reshape((dim, -1)).T

    # Multiply each trial vector by each possible integer up to
    # `search_size`:
    search_ints_rs = search_ints.reshape(-1, 1, 1)
    trial_combs = trials * search_ints_rs

    # Combine trial vectors and their associated scaled vectors:
    trial_combs_all = np.vstack(trial_combs)

    # Find unique vectors. The inverse indices`uinv` indexes
    # the set of unique vectors`u` to generate the original array `pv`:
    uniq, uniq_inv = np.unique(trial_combs_all, axis=0, return_inverse=True)

    # For a given set of (anti-)parallel vectors, we want the smallest, so get
    # their relative magnitudes. This is necessary since `np.unique` does not
    # return vectors sorted in a sensible way if there are negative components.
    # (But we do we have negative components here?)
    uniq_mag = np.sum(uniq**2, axis=1)

    # Get the magnitudes of just the directionally-unique vectors:
    uniq_inv_mag = uniq_mag[uniq_inv]

    # Reshape the magnitudes to allow sorting for a given scale factor:
    uniq_inv_mag_rs = np.reshape(uniq_inv_mag, (search_size, -1))

    # Get the indices which sort the trial vectors
    mag_srt_idx = np.argsort(uniq_inv_mag_rs, axis=0)

    # Reshape the inverse indices
    uniq_inv_rs = np.reshape(uniq_inv, (search_size, -1))

    # Sort the inverse indices by their corresponding vector magnitudes,
    # for each scale factor:
    col_idx = np.tile(np.arange(uniq_inv_rs.shape[1]), (search_size, 1))
    uniq_inv_rs_srt = uniq_inv_rs[mag_srt_idx, col_idx]

    # Only keep inverse indices in first row which are not in any other row.
    # First row indexes lowest magnitude vectors for each scale factor.
    idx = np.setdiff1d(uniq_inv_rs_srt[0], uniq_inv_rs_srt[1:])

    # Sort kept vectors by magnitude
    final_mags = uniq_mag[idx]
    final_mags_idx = np.argsort(final_mags)

    ret = uniq[idx][final_mags_idx]

    return ret


def tile_int_vecs(int_vecs, dim):
    """
    Tile arbitrary-dimension integer vectors such that they occupy a
    half-space.

    """
    # For tiling, there will a total of 2^(`dim` - 1) permutations of the
    # original vector set. (`dim` - 1) since we want to fill a half space.
    i = np.ones(dim - 1, dtype=int)
    t = np.triu(i, k=1) + -1 * np.tril(i)

    # Get permutation of +/- 1 factors to tile initial vectors into half-space
    perms_partial_all = [j for i in t for j in list(permutations(i))]
    perms_partial = np.array(list(set(perms_partial_all)))

    perms_first_col = np.ones((2**(dim - 1) - 1, 1), dtype=int)
    perms_first_row = np.ones((1, dim), dtype=int)
    perms_non_eye = np.hstack([perms_first_col, perms_partial])
    perms = np.vstack([perms_first_row, perms_non_eye])

    perms_rs = perms[:, np.newaxis]
    tiled = int_vecs * perms_rs
    ret = np.vstack(tiled)

    return ret


def find_non_parallel_int_vecs(search_size, dim=3, tile=False):
    """
    Find arbitrary-dimension integer vectors which are non-collinear, whose
    components are less than or equal to a given search size.

    Non-collinear here means no two vectors are related by a scaling factor.
    The zero vector is excluded.

    Parameters
    ----------
    search_size : int
        Positive integer which is the maximum vector component.
    dim : int
        Dimension of vectors to search.
    tile : bool, optional
        If True, the half-space of dimension `dim` is filled with vectors,
        otherwise just the positive vector components are considered. The
        resulting vector set will still contain only non-collinear vectors.

    Returns
    -------
    ndarray of shape (N, `dim`)
        Vectors are not globally ordered.

    Notes
    -----
    Searching for vectors with `search_size` of 100 uses about 9 GB of memory.

    """

    # Find all non-parallel positive integer vectors which have no
    # zero components:
    ret = find_positive_int_vecs(search_size, dim)

    # If requested, tile the vectors such that they occupy a half-space:
    if tile and dim > 1:
        ret = tile_int_vecs(ret, dim)

    # Add in the vectors which are contained within a subspace of dimension
    # (`dim` - 1) on the principle axes. I.e. vectors with zero components:
    if dim > 1:

        # Recurse through each (`dim` - 1) dimension subspace:
        low_dim = dim - 1
        vecs_lower = find_non_parallel_int_vecs(search_size, low_dim, tile)

        # Raise vectors to current dimension with a zero component. The first
        # (`dim` - 1) "prinicple" vectors (of the form [1, 0, ...]) should be
        # considered separately, else they will be repeated.
        principle = np.eye(dim, dtype=int)
        non_prcp = vecs_lower[low_dim:]

        if non_prcp.size:

            edges_shape = (dim, non_prcp.shape[0], non_prcp.shape[1] + 1)
            vecs_edges = np.zeros(edges_shape, dtype=int)
            edges_idx = list(combinations(list(range(dim)), low_dim))

            for i in range(dim):
                vecs_edges[i][:, edges_idx[i]] = non_prcp

            vecs_edges = np.vstack([principle, *vecs_edges])

        else:
            vecs_edges = principle

        ret = np.vstack([vecs_edges, ret])

    return ret


def get_parallel_idx(vecs_a, vecs_b):
    """
    Find which vectors in an array of vectors are (anti-)parallel to which
    vectors in another array of vectors, according to their cross product.

    Parameters
    ----------
    vecs_a : ndarray of shape (3, N)
        Array of column vectors
    vecs_b : ndarray of shape (3, M)
        Array of column vectors

    Returns
    -------
    dict of (int: ndarray of int of shape (P,))
        Returned dict has keys which index `vecs_a` and values which are
        integer arrays which index `vecs_b`.

    """

    ret = {}
    for vec_a_idx, vec_a in enumerate(vecs_a.T):

        vec_a = vec_a[:, np.newaxis]
        vec_a_cross = np.cross(vec_a, vecs_b, axis=0)

        is_parallel = np.all(np.isclose(vec_a_cross, 0.0), axis=0)
        parallel_idx = np.where(is_parallel)[0]

        if parallel_idx.size:
            ret.update({
                vec_a_idx: parallel_idx
            })

    return ret


def get_equal_indices(arr, scale_factors=None):
    """
    Return the indices along the first dimension of an array which index equal
    sub-arrays.

    Parameters
    ----------
    arr : ndarray or list
        Array or list of any shape whose elements along its first dimension are
        compared for equality.
    scale_factors : list of float or list of int, optional
        Multiplicative factors to use when checking for equality between
        subarrays. Each factor is checked independently.

    Returns
    -------
    tuple of dict of int: list of int
        Each tuple item corresponds to a scale factor for which each dict maps
        a subarray index to a list of equivalent subarray indices given that
        scale factor. Length of returned tuple is equal to length of
        `scale_factors` or 1 if `scale_factors` is not specified.

    Notes
    -----
    If we have a scale factor `s` which returns {a: [b, c, ...]}, then the
    inverse scale factor `1/s` will return {b: [a], c: [a], ...}.


    Examples
    --------

    1D examples:

    >>> a = np.array([5, 1, 4, 6, 1, 8, 2, 7, 4, 7])
    >>> get_equal_indices(a)
    ({1: [4], 2: [8], 7: [9]},)

    >>> a = np.array([1, -1, -1, 2])
    >>> get_equal_indices(a, scale_factors=[1, -1, -2, -0.5])
    ({1: [2]}, {0: [1, 2]}, {1: [3], 2: [3]}, {3: [1, 2]})

    2D example:

    >>> a = np.array([[1., 2.], [3., 4.], [-0.4, -0.8]])
    >>> get_equal_indices(a, scale_factors=[-0.4])
    ({0: [2]},)

    """

    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)

    if scale_factors is None:
        scale_factors = [1]

    a_dims = len(arr.shape)
    arr_B = arr[:, np.newaxis]

    sf_shape = tuple([len(scale_factors)] + [1] * (a_dims + 1))
    sf = np.array(scale_factors).reshape(sf_shape)

    bc = np.broadcast_arrays(arr, arr_B, sf)
    c = np.isclose(bc[0], bc[1] * bc[2])

    if a_dims > 1:
        c = np.all(c, axis=tuple(range(3, a_dims + 2)))

    out = ()
    for c_sub in c:

        w2 = np.where(c_sub)
        d = {}
        skip_idx = []

        for i in set(w2[0]):

            if i not in skip_idx:

                row_idx = np.where(w2[0] == i)[0]
                same_idx = list(w2[1][row_idx])

                if i in same_idx:

                    if len(row_idx) == 1:
                        continue

                    elif len(row_idx) > 1:
                        same_idx.remove(i)

                d.update({i: same_idx})
                skip_idx += same_idx

        out += (d,)

    return out



"""Module containing general utility functions."""

import numpy as np


def prt(obj, name, slc=None):
    """
    Print an object, with it's size and dtype if it's an ndarray.

    Helpful for debugging.

    Parameters
    ----------
    obj : object
        Object to print.
    name : str
        Name of object to print.
    slc : tuple of slice object, optional
        If not None, only print the sub array defined by the slice.
    """

    if isinstance(obj, np.ndarray):

        if slc is not None:
            slc_str = ', '.join(['{}:{}:{}'.format(
                s.start or '', s.stop or '', s.step or '') for s in slc])
            slc_str = '[' + slc_str + ']'
        else:
            slc_str = ''

        fmt_str = '{} {} {} {}: \n{}\n'
        fmts = (name, obj.shape, slc_str,
                obj.dtype, obj[slc] if slc is not None else obj)

    elif isinstance(obj, dict):

        fmt_str = ''
        fmts = []
        for key, val in sorted(obj.items()):
            fmt_str += '{}: {}\n'
            fmts.extend([key, val])

    else:

        fmt_str = '{} ({}): \n{}\n'
        fmts = (name, type(obj), obj)

    print(fmt_str.format(*fmts))


def snap_arr(arr, val, tol):
    """Snap array elements within a certain value.

    Parameters
    ----------
    arr : ndarray
    val : float
        Value to snap to.
    tol : float
        Array elements within this tolerance of `val` will be snapped to `val`.

    Returns
    -------
    arr_snapped : ndarray
        Copy of input array, where elements which are close to `val` are set to
        be exactly `val` if within tolerance `tol`.

    """

    arr_snapped = np.copy(arr)
    arr_snapped[abs(arr - val) < tol] = val
    return arr_snapped


def validate_array_args(*args):
    """Validate the shapes of a set of Numpy arrays.

    Parameters
    ----------
    arg : tuple of (str, ndarray, tuple of (str or int))
        Each arg has: the name of a Numpy array, the array itself, and the 
        expected shape of the array. The expected shape is passed as a tuple,
        whose elements maybe integers of strings; if a string, the element
        represent a symbol used to represent the shape in that dimension.

    Notes
    -----
    This function also checks that, if symbols have been repeated in the
    elements of expected array shapes for multiple arrays, those dimension
    sizes are consistent across the different arrays.

    Returns
    -------
    None

    Example
    -------

    """

    symbols = {}
    arr_msg = ('`{}` must be a Numpy array of shape {}')
    multi_msg = ('Inconsistent dimension size for symbol: {}')

    for i in args:

        name = i[0]
        arr = i[1]
        shape = i[2]

        if not isinstance(arr, np.ndarray) or (arr.ndim != len(shape)):
            raise ValueError(arr_msg.format(name, shape))

        for i_idx, i in enumerate(shape):

            if isinstance(i, str):
                if i not in symbols:
                    symbols[i] = arr.shape[i_idx]
                else:
                    if arr.shape[i_idx] != symbols[i]:
                        raise ValueError(multi_msg.format(symbols[i]))
                continue

            if arr.shape[i_idx] != i:
                raise ValueError(arr_msg)


def get_half_space_flip_idx(vecs):
    """Get the indices of vectors that, if multiplied by minus one, would
    ensure no anti-parallel vectors in the set.

    Parameters
    ----------
    vecs : ndarray of inner shape (3, 1)

    Returns
    -------
    flip_idx : ndarray

    Notes
    -----
    For a set of vectors in R3, we might want to map all vectors such that no
    two are anti-parallel. This function returns, for a set of vectors, the
    indices of vectors that require flipping (i.e. multiplying by -1) in order
    to ensure the whole set of vectors is contained within the positive-x half-
    space.

    Specifically, the conditions for flipping are:
        - [x-component < 0] OR
        - [(x-component == 0) and (y-component < 0)] OR
        - [(x-component == 0) and (y-component == 0) and (z-component < 0)]

    """

    in_shp = vecs.shape[-2:]
    if in_shp != (3, 1):
        msg = ('Input array must have inner shape (3, 1), '
               'but has inner shape: {}'.format(in_shp))
        raise ValueError(msg)

    a = vecs[..., 0, 0]
    b = vecs[..., 1, 0]
    c = vecs[..., 2, 0]

    flip_cond = np.logical_or(
        np.logical_or(
            a < 0,
            np.logical_and(np.isclose(a, 0), b < 0),
        ),
        np.logical_and(
            np.logical_and(np.isclose(a, 0), np.isclose(b, 0)),
            c < 0
        )
    )
    flip_idx = np.where(flip_cond)

    return flip_idx



"""Pip installation script."""

from setuptools import find_packages, setup

setup(
    name='vecmaths',
    version="0.1.4",
    description="A collection of (mostly vectorised) maths functions in Python.",
    author='Adam J Plowman, Maria S Yankova',
    packages=find_packages(),
    install_requires=[
        'numpy',
    ],
    classifiers=[
        'Development Status :: 4 - Beta'
    ],
)



"""Module defining functions related to geometry."""
import numpy as np


def get_box_corners(box, origin=None):
    """
    Get all 8 corners of parallelopipeds, each defined by three edge vectors.

    Parameters
    ----------
    box : ndarray of shape (N, 3, 3) or (3, 3)
        Array defining N parallelopipeds, each as three 3D column vectors which
        define the edges of the parallelopipeds.
    origin : ndarray of shape (3, N), optional
        Array defining the N origins of N parallelopipeds as 3D column vectors.

    Returns
    -------
    ndarray of shape (N, 3, 8)
        Returns 8 3D column vectors for each input parallelopiped.

    TODO:
    -   vectorise properly with axis keyword.

    Examples
    --------
    >>> a = np.random.randint(-1, 4, (2, 3, 3))
    >>> a
    [[[ 0  3  1]
      [ 2 -1 -1]
      [ 1  2  0]]

     [[ 0  0  3]
      [ 1  2  0]
      [-1  1 -1]]]
    >>> geometry.get_box_corners(a)
    array([[[ 0.,  0.,  3.,  1.,  3.,  1.,  4.,  4.],
            [ 0.,  2., -1., -1.,  1.,  1., -2.,  0.],
            [ 0.,  1.,  2.,  0.,  3.,  1.,  2.,  3.]],

           [[ 0.,  0.,  0.,  3.,  0.,  3.,  3.,  3.],
            [ 0.,  1.,  2.,  0.,  3.,  1.,  2.,  3.],
            [ 0., -1.,  1., -1.,  0., -2.,  0., -1.]]])

    """

    if box.ndim == 2:
        box = box[np.newaxis]

    num_boxes = box.shape[0]

    if origin is None:
        origin = np.zeros((3, num_boxes), dtype=box.dtype)

    corners = np.zeros((num_boxes, 3, 8), dtype=box.dtype)
    corners[:, :, 1] = box[:, :, 0]
    corners[:, :, 2] = box[:, :, 1]
    corners[:, :, 3] = box[:, :, 2]
    corners[:, :, 4] = box[:, :, 0] + box[:, :, 1]
    corners[:, :, 5] = box[:, :, 0] + box[:, :, 2]
    corners[:, :, 6] = box[:, :, 1] + box[:, :, 2]
    corners[:, :, 7] = box[:, :, 0] + box[:, :, 1] + box[:, :, 2]

    corners += origin.T[:, :, np.newaxis]

    return corners


def get_box_xyz(box, origin=None, faces=False):
    """
    Get coordinates of paths which trace the edges of parallelopipeds
    defined by edge vectors and origins. Useful for plotting parallelopipeds.

    Parameters
    ----------
    box : ndarray of shape (N, 3, 3) or (3, 3)
        Array defining N parallelopipeds, each as three 3D column vectors which
        define the edges of the parallelopipeds.
    origin : ndarray of shape (3, N) or (3,)
        Array defining the N origins of N parallelopipeds as 3D column vectors.
    faces : bool, optional
        If False, returns an array of shape (N, 3, 30) where the coordinates of
        a path tracing the edges of each of N parallelopipeds are returned as
        column 30 vectors.

        If True, returns a dict where the coordinates for
        each face is a key value pair. Keys are like `face01a`, where the
        numbers refer to the column indices of the vectors in the plane of the
        face to plot, the `a` faces intersect the origin and the `b` faces are
        parallel to the `a` faces. Values are arrays of shape (N, 3, 5), which
        define the coordinates of a given face as five 3D column vectors for
        each of the N input parallelopipeds.

    Returns
    -------
    ndarray of shape (N, 3, 30) or dict of str : ndarray of shape (N, 3, 5)
    (see `faces` parameter).

    """

    if box.ndim == 2:
        box = box[np.newaxis]

    num_boxes = box.shape[0]

    if origin is None:
        origin = np.zeros((3, num_boxes), dtype=box.dtype)

    elif origin.ndim == 1:
        origin = origin[:, np.newaxis]

    if origin.shape[1] != box.shape[0]:
        raise ValueError('If `origin` is specified, there must be an origin '
                         'specified for each box.')

    corns = get_box_corners(box, origin=origin)

    face01a = corns[:, :, [0, 1, 4, 2, 0]]
    face01b = corns[:, :, [3, 5, 7, 6, 3]]
    face02a = corns[:, :, [0, 1, 5, 3, 0]]
    face02b = corns[:, :, [2, 4, 7, 6, 2]]
    face12a = corns[:, :, [0, 2, 6, 3, 0]]
    face12b = corns[:, :, [1, 4, 7, 5, 1]]

    coords = [face01a, face01b, face02a, face02b, face12a, face12b]

    if not faces:
        xyz = np.concatenate(coords, axis=2)

    else:
        face_names = ['face01a', 'face01b', 'face02a',
                      'face02b', 'face12a', 'face12b']
        xyz = dict(zip(face_names, coords))

    return xyz
